package info.kofler

import com.zaxxer.hikari.*
import io.ktor.application.*
import io.ktor.features.origin
import io.ktor.html.*
import io.ktor.http.*
import io.ktor.response.*
import io.ktor.request.*
import io.ktor.routing.*
import io.ktor.util.pipeline.*
import kotlinx.html.*
import org.jetbrains.exposed.sql.*
import org.jetbrains.exposed.sql.transactions.*
import org.jetbrains.exposed.sql.transactions.experimental.*
import java.net.URL

// statt einer Datenbank
// val urlMap = mutableMapOf<String, String>()

// HTML-Code für Formular
val shortenForm = """
    <html><body><form action='/shorten' method='post'>
    <p><label for='url'>Geben Sie hier Ihre URL an:</label>
    <p><input type='text' name='url' id='url' size='100'>
    <p><input type='submit' value='OK'>
    </form></body></html>
""".trimIndent()

fun main(args: Array<String>): Unit = io.ktor.server.netty.EngineMain.main(args)

@Suppress("unused") // Referenced in application.conf
@kotlin.jvm.JvmOverloads
fun Application.module(testing: Boolean = false) {
    // Verbindungs-Pool für Datenbank
    val hikConfig = HikariConfig().apply {
         jdbcUrl = "jdbc:mysql://localhost/urldb"
         driverClassName = "com.mysql.cj.jdbc.Driver"
         username = "urluser"
         password = "geheim"
         maximumPoolSize = 12
    }
    Database.connect(HikariDataSource(hikConfig))
    // Tabelle erzeugen, wenn notwendig
    transaction {
        // SchemaUtils.createMissingTablesAndColumns(Urls)
        SchemaUtils.create(Urls)
        // nur zum Testen
        for (url in Url.all())
            println("${url.id}  ${url.shortcode} ${url.fullurl}")
    }

    routing {
        get("/") {
            // Formular anzeigen
            call.respondText(shortenForm, ContentType.Text.Html)
        }
        get("/{id}") {
            // Short-URL suchen, Weiterleitungs-Link anzeigen
            lookupUrl()
        }
        post("/shorten") {
            // Formulardaten verarbeiten, Short-URL erzeugen
            shortenUrl()
        }
    }
}

// Formularauswertung
suspend fun PipelineContext<Unit, ApplicationCall>.shortenUrl() {
    val req = call.request
    val host = "%s://%s:%s".format(req.origin.scheme, req.host(), req.port())
    val data = call.receiveParameters()
    val url = data["url"]
    if (url !is String || url.length > 1000) {
        call.respondText("Invalid data.")
        return
    }
    // simple Validierung
    val result = runCatching { URL(url) }
    if (result.isFailure) {
        call.respondText("Invalid data.")
        return
    }

    suspendedTransactionAsync {
        val chars = "abcdefghijklmnopqrstuvwxyz0123456789"
        var id: String

        // existiert URL schon in DB?
        val dbresult = Url.find { Urls.fullurl.eq(url) }
        if (!dbresult.empty()) {
            // ja, shortcode = id extrahieren
            id = dbresult.first().shortcode
        } else {
            // nein, neuen Eintrag speichern
            do {
                id = (1..6).map { chars.random() }.joinToString("")
            } while (!Url.find { Urls.shortcode.eq(id) }.empty())
            Urls.insert {
                it[fullurl] = url
                it[shortcode] = id
            }
        }

        // Short-Link anzeigen
        val href = "$host/$id"
        call.respondHtml {
            body {
                p { +"Original URL: "
                    a(url) { +url } }
                p { +"Shortened URL: "
                    a(href) { +href } }
                p { +"Shortened URL as text: $href" }
            }
        }
    }
}


suspend fun PipelineContext<Unit, ApplicationCall>.lookupUrl() {
    val id = call.parameters["id"] ?: "invalid data"
    suspendedTransactionAsync {
        val url = Url.find { Urls.shortcode.eq(id) }.firstOrNull()
        if (url is Url) {
            val href = url.fullurl
            call.respondHtml {
                println("now in respondHtml")
                body {
                    p { +"Go to: "
                        a(href) { +href }
                    }
                }
            }
        } else
            call.respondText { "Invalid link." }
    }
}

