package info.kofler.eval

import com.zaxxer.hikari.HikariConfig
import com.zaxxer.hikari.HikariDataSource

import org.jetbrains.exposed.dao.id.*
import org.jetbrains.exposed.sql.*
import org.jetbrains.exposed.sql.`java-time`.*
import org.jetbrains.exposed.sql.transactions.transaction
import org.mindrot.jbcrypt.BCrypt
import java.math.BigDecimal
import java.time.LocalDate

// Datenbankcode
class Db {
    // private val db: Database

    init {
        // H2
        // Database.connect("jdbc:h2:./myh2", "org.h2.Driver")

        // MySQL oder MariaDB + Connection Pooling
        val config = HikariConfig().apply {
            jdbcUrl         = "jdbc:mysql://localhost/evaldb"
            driverClassName = "com.mysql.cj.jdbc.Driver"
            username        = "evaluser"
            password        = "geheim"
            maximumPoolSize = 12
        }
        val dataSource = HikariDataSource(config)
        Database.connect(dataSource)

        // Tabellen erzeugen (falls erforderlich)
        transaction {
            SchemaUtils.createMissingTablesAndColumns(
                Students, Courses, Attendees, Questions, Evaluations,
                EvalMix, Invitations, Answers)
        }
        // Testdatensätze erzeugen
        createTestData()
    }

    // Musterdatensätze erzeugen
    private fun createTestData() {
        transaction {
            // wenn es schon Datensätze gibt: abbrechen
            if (Students.selectAll().count() > 0)
                return@transaction

            // 10 Musterstudenten
            val students = listOf(
                "Anzenberger Astrid",
                "Endhofer Erich",
                "Gstettner Gabi",
                "Huber Harald",
                "Lackner Ludwig",
                "Müller Maria",
                "Prasser Petra",
                "Reiter Ronja",
                "Schmidt Sonja",
                "Wiesenhofer Wilhelm"
            )
            val studentIds = mutableListOf<EntityID<Int>>()
            for (s in students) {
                studentIds += Students.insert { newStudent ->
                    newStudent[name] = s
                    newStudent[studentNumber] = "2020" + (100_000..999_999).random()
                    newStudent[onetimekey] = (0..999_999).random()
                }[Students.id]
            }
            // 4 Lehrveranstaltungen
            val courses = listOf("Software Development", "Datenbankdesign", "Datenstrukturen und Algorithmen", "Betriebssysteme")
            val courseIds = mutableListOf<EntityID<Int>>()
            for (c in courses) {
                courseIds += Courses.insert { newCourse ->
                    newCourse[name] = c
                    newCourse[time] = "Wintersemester 2020"
                }[Courses.id]
            }
            // alle Studenten nehmen an allen LV teil
            for (cid in courseIds)
                for(sid in studentIds)
                    Attendees.insert { newattendee ->
                        newattendee[courseId] = cid
                        newattendee[studentId] = sid
                    }
            // Fragenpool
            val questionPool = listOf("Unterricht ist verständlich",
                "Vortragende/r ist fachlich kompetent",
                "Vortragende/r ist engagiert/motiviert",
                "Beurteilungskriterien sind nachvollziehbar",
                "Folien und sonstige Materialien sind geeignet",
                "Fach ist für mich relevant/wichtig",
                "TXT: Sonstige Anmerkungen")
            val questionIds = mutableListOf<EntityID<Int>>()
            for (q in questionPool) {
                var qtxt = q
                var type = "range"
                if (q.startsWith("TXT: ")) {
                    qtxt = q.substring(4)
                    type = "text"
                }
                questionIds += Questions.insert { newquestion ->
                    newquestion[text] = qtxt
                    newquestion[qtype] = type
                }[Questions.id]
            }

            // aktive Evaluierungen: für alle Lehrveranstaltungen
            val evalIds = mutableListOf<EntityID<Int>>()
            for (cid in courseIds) {
                evalIds += Evaluations.insert { neweval ->
                    neweval[courseId] = cid
                    neweval[startDate] = LocalDate.now()
                    neweval[endDate] = LocalDate.now().plusDays(28)
                }[Evaluations.id]
            }

            // bei allen Evaluierungen einfach alle Fragen aus dem Fragenpool berücksichtigen
            for(e in evalIds) {
                var seq = 0
                for (q in questionIds)
                    EvalMix.insert { mix ->
                        mix[evaluationId] = e
                        mix[questionId] = q
                        mix[sequence] = seq
                        seq += 1
                    }
            }

            // alle Studenten zu allen Evaluierungen einladen
            for(e in evalIds) {
                for(s in studentIds) {
                    Invitations.insert { invit ->
                        invit[studentId] = s
                        invit[evaluationId] = e
                        var code = ""
                        for (i in 1..16)
                            code += "abcdefghijklmnopqrstuvwxyz".random()
                        invit[key] = code
                    }
                }
            }

            // zufällig einen Teil der Evaluationen durchführen
            val activeInvit = Invitations.select {
                Invitations.active.eq(true)
            }
            for (invit in activeInvit) {
                if ((0..100).random() >= 50)
                    continue  // ignorieren
                val stud = invit[Invitations.studentId]
                val eval = invit[Invitations.evaluationId]

                // alle Fragen durchlaufen + mit Zufallswerten beantworten
                val questions = (EvalMix innerJoin Questions)
                    .slice(Questions.id, Questions.text, Questions.qtype)
                    .select {
                        EvalMix.evaluationId.eq(eval)
                    }
                for (quest in questions) {
                    var text: String? = null
                    var nmb: Int? = null
                    if (quest[Questions.qtype] == "text")
                        text = "lorem ipsum " + (1..1000).random()
                    else
                        nmb = (40..100).random()
                    Answers.insert { newanswer ->
                        newanswer[evaluationId] = eval
                        newanswer[studentId] = stud
                        newanswer[questionId] = quest[Questions.id]
                        newanswer[answerText] = text
                        newanswer[answerNumber] = nmb
                    }
                }

                // als erledigt markieren
                Invitations.update( { Invitations.studentId.eq(stud) and
                        Invitations.evaluationId.eq(eval) }) { row ->
                    row[active] = false
                }
            }

        } // transaction
    }

    // alle offenen Einladungen
    fun invitations() : List<Invitation> {
        val result = mutableListOf<Invitation>()
        transaction {
            val query = (Invitations innerJoin Students
                    innerJoin Evaluations innerJoin Courses)
                .slice(Invitations.key, Courses.name, Courses.time, Students.name)
                .select {
                    Invitations.active.eq(true)
                }
                .orderBy(Courses.name)
                .orderBy(Students.name)
            for (row in query) {
                val key = row[Invitations.key]
                val course = "${row[Courses.name]} (${row[Courses.time]})"
                val student = row[Students.name]
                result += Invitation(key, course, student)
            }
        }
        return result
    }

    // die letzten 20 Evaluierungen, die neusten zuerst
    fun evaluations() : List<Evaluation> {
        val result = mutableListOf<Evaluation>()
        transaction {
            houseKeeping()  // evt.  abgelaufene Eval. deaktivieren

            // Anzahl der Einladungen/Teilnehmer für alle aktiven Evaluierungen
            val noOfInvit = invitationsCount()
            val noOfPart = participantsCount()

            // Evaluierungen inklusive Name der Lehrveranstaltung
            val query = (Evaluations innerJoin Courses)
                    .slice(Evaluations.id, Evaluations.active,
                            Courses.name, Courses.time)
                    .selectAll()
                    .orderBy(Evaluations.startDate to SortOrder.DESC)
                    .limit(20)
            for (row in query) {
                val id = row[Evaluations.id].value
                val course = "${row[Courses.name]} (${row[Courses.time]})"
                val active = row[Evaluations.active]
                val participants = noOfPart[id] ?: 0
                val invitations = noOfInvit[id] ?: 0
                result += Evaluation(id, course, active, participants, invitations)
            }
        }
        return result
    }

    // die letzten 20 Evaluierungen für einen bestimmten Studenten
    fun evaluations(studentNo: String) : List<EvaluationMeta> {
        val result = mutableListOf<EvaluationMeta>()
        transaction {
            val query = (Evaluations innerJoin Courses innerJoin Invitations innerJoin Students)
                .slice(Evaluations.id,
                       Evaluations.active,
                       Courses.name,
                       Courses.time,
                       Evaluations.endDate,
                       Invitations.active)
                .select { Students.studentNumber.eq(studentNo) }
                .orderBy(Evaluations.endDate to SortOrder.DESC)
                .limit(20)
            for (row in query) {
                val id = row[Evaluations.id].value
                val cName = row[Courses.name]
                val cTime = row[Courses.time]
                val eDate = row[Evaluations.endDate].toString()
                val evalActive = row[Evaluations.active]
                val invitOpen = row[Invitations.active]
                result += EvaluationMeta(id, cName, cTime,  eDate, evalActive, invitOpen)
            }
        }
        return result
    }

    // Ergebnisse einer Evaluierung liefern
    fun evaluationResult(evalId: Int) : EvaluationDetails {
        val items = mutableListOf<EvaluationItem>()
        val course = courseName(evalId)
        val invitations = invitationsCount(evalId)
        val participants = participantsCount(evalId)
        transaction {
            // alle Fragen der konkreten Evaluation
            val questions = (Evaluations innerJoin EvalMix innerJoin Questions)
                    .slice(Questions.columns)
                    .select {
                        EvalMix.evaluationId.eq(evalId)
                    }
                    .orderBy(EvalMix.sequence)
            // Schleife über alle Fragen, Ergebnisse sammeln
            for (q in questions) {
                val qtype = q[Questions.qtype]
                val qtxt = q[Questions.text]
                val qid = q[Questions.id].value
                if (qtype == "range") {
                    val avg = (Answers)
                            .slice(Answers.answerNumber.avg())
                            .select {
                                Answers.evaluationId.eq(evalId) and
                                        Answers.questionId.eq(q[Questions.id])
                            }
                            .first()[Answers.answerNumber.avg()]
                    if (avg is BigDecimal)
                        items += EvaluationItem(qid, qtxt, qtype, avg.toInt())
                } else if (qtype == "text") {
                    val txt = (Answers)
                            .slice(Answers.answerText)
                            .select {
                                Answers.evaluationId.eq(evalId) and
                                        Answers.questionId.eq(q[Questions.id]) and
                                        Answers.answerText.isNotNull()
                            }
                            .map { row -> row[Answers.answerText] }
                            .joinToString(separator = "|")
                    if (txt.length > 2)
                        items += EvaluationItem(qid, qtxt, qtype, null, txt)
                }
            }
        }
        return EvaluationDetails(evalId, course, items, null, null, invitations, participants)
    }

    // liefert die Fragestellung für eine Evaluierung,
    // wobei der Invitations-Key als Parameter übergeben wird
    fun questions(invitKey: String) : EvaluationDetails? {
        val evalId = evalId(invitKey) ?: return null
        val course = courseName(evalId)

        val items = mutableListOf<EvaluationItem>()
        transaction {
            val query = (Evaluations innerJoin EvalMix innerJoin Questions)
                    .slice(Questions.columns)
                    .select { Evaluations.id.eq(evalId) }
            for (q in query)
                items += EvaluationItem(q[Questions.id].value,
                        q[Questions.text],
                        q[Questions.qtype])
        }
        return EvaluationDetails(evalId, course, items)
    }
    // wie oben, aber für eine bestimmten evalId
    fun questions(evalId: Int) : EvaluationDetails {
        val course = courseName(evalId)
        val items = mutableListOf<EvaluationItem>()
        transaction {
            val query = (Evaluations innerJoin EvalMix innerJoin Questions)
                    .slice(Questions.columns)
                    .select { Evaluations.id.eq(evalId) }
            for (q in query)
                items += EvaluationItem(q[Questions.id].value,
                        q[Questions.text],
                        q[Questions.qtype])
        }
        return EvaluationDetails(evalId, course, items)
    }

    // Daten einer Evaluierung speichern
    // in data muss entweder eine gültige studentId oder studentNo
    // angegeben werden
    // Ergebnisdatentyp: Boolean
    fun saveEvaluation(data: EvaluationDetails) =
        // transaction-Lambda-Ausdruck liefert true/false zurück,
        // je nachdem, ob Speichern möglich ist oder nicht
        transaction {
            println("xxx $data")

            val evalId = data.evalId
            val studId = data.studentId                 // wenn null: studentNo auswerten
                ?: studentIdfromStudentNo(data.studentNo)   // wenn auch null: Abbruch
                ?: return@transaction false

            // Valididätsprüfung
            val valid  = (Invitations innerJoin Evaluations)
                .select {
                    Invitations.studentId.eq(studId) and
                            Invitations.evaluationId.eq(evalId) and
                            Invitations.active.eq(true) and
                            Evaluations.active.eq(true) and
                            Evaluations.startDate.lessEq(LocalDate.now()) and
                            Evaluations.endDate.greaterEq(LocalDate.now())
                }.count() == 1L
            if (!valid)
                return@transaction false

            // Fragetypen ermitteln
            val qtypes =
              (Evaluations innerJoin EvalMix innerJoin Questions)
                .slice(Questions.id, Questions.qtype)
                .select { Evaluations.id.eq(data.evalId) }
                .associate { it[Questions.id].value to it[Questions.qtype]}

            /*  // gleichwertig, aber umständlicher
            val qtypes = mutableMapOf<Int, String>()
            val query = (Evaluations innerJoin EvalMix innerJoin Questions)
                .slice(Questions.id, Questions.qtype)
                .select { Evaluations.id.eq(data.evalId) }
            for (row in query)
                qtypes[row[Questions.id].value] = row[Questions.qtype]
            */

            // Schleife über Evaluierungsdaten
            for (itm in data.items) {
                println("xxx item = $itm")
                val qtype = qtypes.getOrDefault(itm.questId, "invalid")
                if (qtype == "range") {
                    // Prozentzahl speichern
                    val pct = itm.nmb ?: itm.answer?.toIntOrNull() ?: continue
                    if (pct in 0..100) {
                        Answers.insert { newanswer ->
                            newanswer[studentId] = EntityID(studId, Students)
                            newanswer[evaluationId] = EntityID(data.evalId, Evaluations)
                            newanswer[questionId] = EntityID(itm.questId, Questions)
                            newanswer[answerNumber] = pct
                        }
                    }
                } else if (qtype == "text" ) {
                    // Text speichern
                    val answer = itm.answer ?: continue
                    if(answer.isNotBlank())
                        Answers.insert { newanswer ->
                            newanswer[studentId] = EntityID(studId, Students)
                            newanswer[evaluationId] = EntityID(data.evalId, Evaluations)
                            newanswer[questionId] = EntityID(itm.questId, Questions)
                            newanswer[answerText] = answer
                    }
                }
            }

            // Invitation-Link deaktivieren
            Invitations.update( // Bedingung
                { Invitations.studentId.eq(studId) and
                        Invitations.evaluationId.eq(evalId) } )
            { row -> row[active] = false }  // Änderung

            // vollständige Evaluationen + abgelaufene Einladungen
            // deaktivieren
            houseKeeping()

            // alles bestens, true zurückgeben
            true
        }


    // deaktiviert abgelaufene Evaluationen und Einladungen
    private fun houseKeeping() {
        transaction {
            // zeitlich abgelaufene Evaluierungen
            Evaluations.update( {
               Evaluations.endDate.less(LocalDate.now())
            }) {
                row -> row[active] = false
            }

            // alle noch aktiven Eval. untersuchen
            val noOfInvit = invitationsCount()
            val noOfPart = participantsCount()
            val activeEval = Evaluations.slice(Evaluations.id)
                    .select { Evaluations.active.eq(true) }
            for (eval in activeEval) {
                val evalId = eval[Evaluations.id].value
                val participants = noOfPart[evalId] ?: continue
                val invitations = noOfInvit[evalId] ?: continue
                // deaktivieren, wenn alle teilgenommen haben
                if (participants >= invitations)
                    Evaluations.update( {
                        Evaluations.id.eq(evalId)
                    }) {
                        row -> row[active] = false
                    }
            }

            // alle Einladungen inaktiver Evaluationen deaktivieren
            val inactive = (Invitations innerJoin Evaluations)
                    .slice(Invitations.evaluationId)
                    .select {
                        Evaluations.active.eq(false) and Invitations.active.eq(true)
                    }
                    .withDistinct(true)
                    .map { row -> row[Invitations.evaluationId] }
            if (inactive.isNotEmpty())
                (Invitations).update( {
                    Invitations.evaluationId.inList(inactive)
                } )
                { row -> row[active] = false }
        }
    }

    // Hilfsfunktionen
    // testet, ob Evaluierungs-Key existiert und die dazugehörende
    // Evaluierung noch aktiv ist; liefert die Evaluierungs-ID oder null zurück
    fun evalId(key: String) : Int? {
        var id: Int? = null
        transaction {
            val query = (Evaluations innerJoin Invitations)
                    .slice(Evaluations.id)
                    .select { Invitations.key.eq(key) and
                            Invitations.active.eq(true) and
                            Evaluations.active.eq(true) and
                            Evaluations.startDate.lessEq(LocalDate.now()) and
                            Evaluations.endDate.greaterEq(LocalDate.now())
                    }
            val firstrow = query.firstOrNull()
            if (firstrow is ResultRow)
              id = firstrow[Evaluations.id].value
        }
        return id
    }

    // ermittelt Studenten-ID aus Invitation-Key
    // Ergebnistyp Int?
    fun studentIdfromKey(key: String) =
        transaction {
            return@transaction  Invitations
                .slice(Invitations.studentId)
                .select { Invitations.key.eq(key) }
                .firstOrNull()
                ?.get(Invitations.studentId)
                ?.value
        }

    // ermittelt Studenten-ID aus studentNo (Matrikelnr.)
    // Ergebnistyp Int?
    private fun studentIdfromStudentNo(studentNo: String?) =
        if (studentNo == null || studentNo.isBlank())
            null
        else
            transaction {
                return@transaction Students
                    .slice(Students.id)
                    .select { Students.studentNumber.eq(studentNo) }
                    .firstOrNull()
                    ?.get(Students.id)
                    ?.value
            }

    // ermittelt Name und Zeit einer Lehrveranstaltung aus der Evaluierungs-ID
    // liefert leere Zeichenkette, wenn die ID ungültig ist
    private fun courseName(evalId: Int) : String {
        var course = ""
        transaction {
            val row = (Evaluations innerJoin Courses)
                    .slice(Courses.name, Courses.time)
                    .select { Evaluations.id.eq(evalId) }
                    .firstOrNull()
            if (row is ResultRow)
                course = "${row[Courses.name]} (${row[Courses.time]})"
        }
        return course
    }

    // liefert Anz. Einladungen aller aktiven bzw. einer bestimmten Evaluierungen
    private fun invitationsCount() : Map<Int, Int> {
        val result = mutableMapOf<Int, Int>()
        transaction {
            val query = (Evaluations innerJoin Invitations)
                    .slice(Evaluations.id, Invitations.studentId.count())
                    .select { Evaluations.active.eq(true) }
                    .groupBy(Evaluations.id)
            for (row in query)
                result[row[Evaluations.id].value] =
                        row[Invitations.studentId.count()].toInt()
        }
        return result
    }
    private fun invitationsCount(evalId: Int) : Int {
        return transaction {
            val query = (Evaluations innerJoin Invitations)
                    .slice(Evaluations.id, Invitations.studentId.count())
                    .select { Evaluations.id.eq(evalId) }
                    .groupBy(Evaluations.id)
            val row = query.firstOrNull() ?: return@transaction 0
            return@transaction row[Invitations.studentId.count()].toInt()
        }
    }

    // liefert Anz. Teilnahmen aller aktiven bzw. einer bestimmten Evaluierungen
    private fun participantsCount() : Map<Int, Int> {
        val result = mutableMapOf<Int, Int>()
        transaction {
            val query = (Evaluations innerJoin Answers)
                    .slice(Evaluations.id, Answers.studentId.countDistinct())
                    .select { Evaluations.active.eq(true) }
                    .withDistinct(true)
                    .groupBy(Evaluations.id)
            for (row in query)
                result[row[Evaluations.id].value] =
                        row[Answers.studentId.countDistinct()].toInt()
        }
        return result
    }
    private fun participantsCount(evalId: Int) =
        transaction {
            val query = (Evaluations innerJoin Answers)
                    .slice(Evaluations.id, Answers.studentId.countDistinct())
                    .select { Evaluations.id.eq(evalId) }
                    .withDistinct(true)
                    .groupBy(Evaluations.id)
            val row = query.firstOrNull() ?: return@transaction 0
            return@transaction row[Answers.studentId.countDistinct()].toInt()
        }


    // liefert Students.onetimekey aus Invitation-Key
    fun oneTimeKey(invitKey: String) : Int? {
        return transaction {
            return@transaction (Invitations innerJoin Students)
                    .slice(Students.onetimekey)
                    .select { Invitations.key.eq(invitKey) }
                    .firstOrNull()
                    ?.get(Students.onetimekey)
        }
    }
    // liefert Students.studentNumber aus Invitation-Key
    fun studentNumber(invitKey: String) : String? {
        return transaction {
            return@transaction (Invitations innerJoin Students)
                    .slice(Students.studentNumber)
                    .select { Invitations.key.eq(invitKey) }
                    .firstOrNull()
                    ?.get(Students.studentNumber)
        }
    }

    // One-Time-Login, falls erfolgreich, wird Hash von
    // appkey gespeichert; liefert true/false zurück
    fun oneTimeLogin(data: LoginData) =
        transaction {
            val onetimekey = data.onetimekey.toIntOrNull() ?: -1
            val hash = BCrypt.hashpw(data.appkey, BCrypt.gensalt())
            val id = Students
                .slice(Students.id)
                .select {Students.studentNumber.eq(data.studentNo) and
                        Students.onetimekey.eq(onetimekey)}
                .firstOrNull()
                ?.get(Students.id)
            if (id is EntityID<Int>) {
                Students.update( {
                    Students.id.eq(id)
                }) { row ->
                    row[Students.onetimekey] = null
                    row[Students.apphash] = hash
                }
                return@transaction true
            }
            return@transaction  false
        }

    // überprüft den aus StudentNo + Passwort bestehenden Login
    // liefert Students.id oder null zurück
    fun standardLogin(studentNo: String, password: String) =
        transaction {
            // gibt es einen passenden Datensatz?
            val row = Students
                .slice(Students.id, Students.apphash)
                .select {
                    Students.studentNumber.eq(studentNo) and
                            Students.apphash.isNotNull()
                }
                .firstOrNull()
                ?: return@transaction null

            val hash = row[Students.apphash]
            if (BCrypt.checkpw(password, hash))
                row[Students.id].value
            else
                null
        }

}






// Singletons für alle Tabellen (Datenbank-Schema)
object Students : IntIdTable() {
    val name   = varchar("name", length = 100).index()
    val studentNumber = varchar("studentNumber", length = 12).uniqueIndex()
    val onetimekey = integer("onetimekey").nullable()
    val apphash = varchar("appkey", length = 255).nullable()
}
object Courses : IntIdTable() {
    val name = varchar("name", length = 200).index()
    val time = varchar("time", length = 100).index()
}
object Attendees : Table() {
    val courseId = reference("courseId", Courses,
        onDelete = ReferenceOption.CASCADE)
    val studentId =reference("studentId", Students,
        onDelete = ReferenceOption.CASCADE)
    override val primaryKey = PrimaryKey(courseId, studentId)
}
object Questions : IntIdTable() {
    val text = varchar("text", length = 250)
    val qtype = varchar("qtype", length = 20)  // "range" oder "text"
}
object Evaluations: IntIdTable() {
    val courseId = reference("courseId", Courses,
        onDelete = ReferenceOption.CASCADE)
    val startDate = date("startDate")
    val endDate = date("endDate")
    val active = bool("active").default(true)
}
object EvalMix: Table() {
    val evaluationId = reference("evaluationId", Evaluations,
        onDelete = ReferenceOption.CASCADE)
    val questionId =reference("questionId", Questions,
        onDelete = ReferenceOption.CASCADE)
    val sequence = integer("sequence")
    override val primaryKey = PrimaryKey(evaluationId, questionId)
}
object Invitations: Table() {
    val evaluationId = reference("evaluationId", Evaluations,
        onDelete = ReferenceOption.CASCADE)
    val studentId =reference("studentId", Students,
        onDelete = ReferenceOption.CASCADE)
    val key = varchar("key", 16)
    val active = bool("active").default(true)
    override val primaryKey = PrimaryKey(evaluationId, studentId)
}
object Answers: Table() {
    val evaluationId = reference("evaluationId", Evaluations,
        onDelete = ReferenceOption.CASCADE)
    val studentId =reference("studentId", Students,
        onDelete = ReferenceOption.CASCADE)
    val questionId =reference("questionId", Questions,
        onDelete = ReferenceOption.CASCADE)
    val answerText = varchar("answerText", 1000).nullable()
    val answerNumber = integer("answerNumber").nullable()
    override val primaryKey = PrimaryKey(evaluationId, studentId, questionId)
}

