package info.kofler.eval

import io.ktor.application.*
import io.ktor.routing.*
import io.ktor.html.*
import io.ktor.http.Parameters
import io.ktor.http.content.resource
import io.ktor.http.content.static
import io.ktor.request.receiveParameters
import io.ktor.response.respondText
import kotlinx.html.*

fun Routing.webrouting() {
   startpage()                 // Startseite
   evaluations()               // Liste der 20 aktuellsten Evaluierungen
   evaluationDetails()         // Detailergebnisse einer Evaluierung
   invitations()               // Liste aller noch offenen Evaluierungs-Einladungen
   evaluationForm()            // Seite zur Durchführung einer Evaluierung
   processEvaluationData()    // Evaluierungsergebnisse (Formulardaten) verarbeiten
   css()
}

// Startseite
fun Routing.startpage() {
   get("/") {
      call.respondHtml {
         myhead()
         body {
            h1 { +"Startseite" }
            p {
               a(href = "results") {
                  +"Evaluierungsergebnisse"
               }
            }
            p {
               a(href = "invitations") {
                  +"Einladungs-Links für die Evaluierung"
               }
            }
         }
      }
   }
}

// Liste der zuletzt durchgeführten Evaluierungen zeigen
fun Routing.evaluations() {
   get("/results") {
      call.respondHtml {
         myhead()
         body {
            h1 { +"Evaluierungen" }
            ul {
               for (e in db.evaluations()) {
                  li {
                     p {
                        a(href = "evaluation/${e.id}") {
                           +e.course
                        }
                        if (e.active) {
                           +" (noch aktiv, bisher teilgenommen:  ${e.participants} von ${e.invitations})"
                        }
                     }
                  }
               }
            }
         }
      }
   }
}

// Detailergebnisse einer Evaluierung
fun Routing.evaluationDetails() {
   get("/evaluation/{id}") {
      val id = call.parameters["id"]?.toInt()
      if (id is Int) {
         val results = db.evaluationResult(id)
         call.respondHtml {
            myhead()
            body {
               h1 { +results.course }
               p { +"${results.participants} von ${results.invitations} "
                  +"Personen haben an der Evaluierung teilgenommen" }
               var no = 1
               for (itm in results.items) {
                  p {
                     +"$no. ${itm.question}: "
                     if (itm.answer is String) {
                        for (answer in itm.answer.split("|"))
                           ul { li { +answer } }
                     } else if (itm.nmb is Int)
                        +"${itm.nmb}%"
                  }
                  no += 1
               }
            }
         }

      } else
         call.respondText { "Invalid id" }
   }
}

// Sammlung aller aktiven Evaluierungs-Links für Schüler/Studenten
fun Routing.invitations() {
   get("/invitations") {
      call.respondHtml {
         myhead()
         body {
            h1 { +"Offene Einladungen zu Evaluierungen" }
            ul {
               for (i in db.invitations()) {
                  li {
                     p {
                        a(href = "evalform/${i.key}") {
                           +"${i.course} / ${i.student}"
                        }
                     }
                  }
               }
            }
         }
      }
   }
}

// Evaluierungsseite anzeigen
fun Routing.evaluationForm() {
   get("/evalform/{key}") {
      val key = call.parameters["key"] ?: "invalid key"
      if (db.evalId(key) is Int) {
         val eval = db.questions(key)!!  // key ist ja überprüft
         call.respondHtml {
            myhead()
            body {
               h1 { +"Evaluierung für ${eval.course}" }
               evalForm(key, eval)  // eigentliches Formular aufbauen
               loginData(key)       // Login-Daten für App
            }
         }
      } else
         call.respondText { "Invalid key" }
   }
}

// HTML-Formular für die Evaluierung
fun BODY.evalForm(invitKey: String, eval: EvaluationDetails) {
   form(action = "/processeval/$invitKey", method = FormMethod.post) {
      table {
         for (qst in eval.items) {
            tr {
               td { +qst.question }
               if (qst.qtype == "text")
                  td {
                     unsafe {
                        +"""<textarea name='${qst.questId}' 
                            rows='3' cols='40'></textarea>"""
                     }
                  }
               else if (qst.qtype == "range")
                  td { radio7("${qst.questId}") } // sieben Radio-Buttons
            }
         }
         // Letzte Zeile: Senden-Button
         tr {
            td { }
            td { submitInput() }
         }
      }
   }
}

// One-time-key für App-Login anzeigen (falls verfügbar)
fun BODY.loginData(invitKey: String) {
   val key = db.oneTimeKey(invitKey)
   if (key is Int) {
      val studentNmb = db.studentNumber(invitKey)
      p {
         +"""Alternativ können Sie die Evaluierung auch auf Ihrem 
                Smartphone vornehmen. Laden Sie die Evaluierungs-App 
                herunter und melden Sie sich mit Ihrer Studenten-ID 
                ($studentNmb) und dem folgenden Code an: $key"""
      }
   }
}


// Gruppe aus sieben Radio-Buttons
fun TD.radio7(name: String) {
   var html = "schlecht "
   for (i in 0..6) {
      val perct = (100.001 / 6 * i).toInt()
      html += "<input type='radio' name='$name' value='$perct'>"
   }
   html += " gut"
   unsafe { +html }
}

// Formulardaten einer Evaluierung auswerten
fun Routing.processEvaluationData() {
   post("/processeval/{key}") {
      val key = call.parameters["key"] ?: "invalid key"
      val evalId = db.evalId(key)
      val studentId = db.studentIdfromKey(key)
      if (evalId is Int && studentId is Int) {
         val ok = processEval(evalId, studentId, call.receiveParameters())
         call.respondHtml {
            myhead()
            body {
               h1 { +"Evaluierung speichern" }
               if (ok)
                  p { +"Ihre Daten wurden gespeichert. Vielen Dank für die Evaluierung!" }
               else
                  p { +"Es ist leider ein Fehler aufgetreten. Bitte füllen Sie das Formular nochmals aus." }
            }
         }
      } else
         call.respondText { "Invalid key" }
   }
}

private fun processEval(evalId: Int, studentId: Int, data: Parameters): Boolean {
   if (data.entries().isEmpty())
      return false

   // Formulardaten in EvaluationDetails-Objekt zusammenstellen ...
   val answers = mutableListOf<EvaluationItem>()
   for ((paraname, valList) in data.entries()) {
      val qid = paraname.toIntOrNull() ?: continue
      val value = valList.firstOrNull() ?: continue
      answers += EvaluationItem(qid, answer = value)
   }
   val evalDetails = EvaluationDetails(evalId, "", answers, studentId)
   // ... und in der Datenbank speichern
   return db.saveEvaluation(evalDetails)
}

// statische Inhalte (CSS)
fun Routing.css() {
   static("/static") {
      resource("eval.css", resource = "eval.css")
   }
}

// gemeinsamer Header für alle HTML-Seiten
fun HTML.myhead() {
   head {
      title { +"Evaluierungssystem" }
      link(href = "/static/eval.css", type = "text/css", rel = "stylesheet")
   }
}
